function [anomalies] = detect_anomalies(data, threshold)

% [anomalies] = detect_anomalies(data, threshold)
% 
% This function detect anomalies in the dataset (feature matrix) stored in
% the input variable "data" (rows: observations, cols: features) using the
% multivariate Gaussian distribution.
% 
% This anomaly detection algorithm is based on the computation of sample 
% mean and variance for each feature and estimation of probability density 
% function (p). The features with p < "threshold" are marked as anomalies.
% 
% INPUTS:
% data          - input feature matrix (rows: observs, cols: features)
% threshold     - input threshold [defaulr: 0.05] => 95% probability
% 
% OUTPUTS:
% anomalies     - output anomalies detections (index vector)
%
% *Visualization is performed only for 2D (2 features) datasets
% **Visualization assumes the data are positive with comparable ranges
%
%
%
% --
% ing. Zoltn Gal
% xgalaz00@stud.feec.vutbr.cz       
% 
% Department of Telecommunications
% Faculty of Electrical Engineering and Communication
% Brno University of Technology

%% Paths and variables
if ((nargin < 2) || isempty(threshold))
    threshold = 0.05;
end

%% Estimate the parameters of the Gaussian distribution    
[mu, sigma_2] = estimate_gaussian(data);

%% Compute the distribution probability density function
p_density = multivariate_gaussian(data, mu, sigma_2);

%% Detect anomalies (data with: p < threshold)
anomalies = find(p_density < threshold);

%% Visualize the dataset/distribution (2D data)
if (size(data, 2) == 2)
    
    % create the grid
    [x, y] = meshgrid(0:.5:max(max(data))*1.5);
    
    % Estimate the probability density function
    Z = multivariate_gaussian([x(:) y(:)], mu, sigma_2);
    Z = reshape(Z, size(x));
    
    % Plot the data
    plot(data(:, 1), data(:, 2), 'bx');
    hold on;
    grid on;
    
    % Check for infinity values
    if (sum(isinf(Z)) == 0)
        contour(x, y, Z, 10.^(-20:3:0)');
    end
    
    % Mark the anomalies
    plot(data(anomalies, 1), data(anomalies, 2), ...
        'ro', 'LineWidth', 2, 'MarkerSize', 10);
    
    % Set the title/axes
    xlabel('feature x1');
    ylabel('feature x2');
    title(['Detected: ' num2str(length(anomalies)) ...
        ' anomalies, threshold: p < ' num2str(threshold)]);
    
    hold off;
end